import json


def clean_empty(d):
    """Recursively remove empty lists, empty dicts, or None elements from a dictionary."""
    if not isinstance(d, (dict, list)):
        return d
    if isinstance(d, list):
        return [v for v in (clean_empty(v) for v in d) if v != [] and v is not None]
    return {
        k: v
        for k, v in ((k, clean_empty(v)) for k, v in d.items())
        if v is not None and v != {}
    }


def custom_serializer(obj):
    """Fallback function to convert unserializable objects."""
    if hasattr(obj, "__dict__"):
        # Attempt to serialize custom objects by their __dict__ attribute.
        return clean_empty(obj.__dict__)
    else:
        # For other types, just convert to string
        return str(obj)


def non_serialize(value):
    return f"<<non-serializable: {type(value).__qualname__}>>"


def write_to_json(value, *args, **kwargs):
    cleaned_object = clean_empty(value)
    try:
        return json.dump(cleaned_object, *args, **kwargs)
    except Exception as e:
        print(e)
        try:
            return json.dump(cleaned_object, *args, default=custom_serializer, **kwargs)
        except Exception as e:
            print(e)
            try:
                return json.dump(cleaned_object, *args, default=non_serialize, **kwargs)
            except Exception as e:
                print(e)
                return value


def serialise_to_json(value) -> str:
    cleaned_object = clean_empty(value)
    try:
        return json.dumps(cleaned_object)
    except Exception as e:
        print(e)
        try:
            return json.dumps(cleaned_object, default=custom_serializer)
        except Exception as e:
            print(e)
            try:
                return json.dumps(cleaned_object, default=non_serialize)
            except Exception as e:
                print(e)
                return value


__all__ = [
  'serialise_to_json',
  'write_to_json',
]



from .chat_completion import OpenAIChatCompletionFactory
from .images_response import OpenAIImagesResponseFactory
from .api_client import MockOpenAISyncAPIClient, MockOpenAIAsyncAPIClient


__all__ = [
  'OpenAIChatCompletionFactory',
  'OpenAIImagesResponseFactory',
  'MockOpenAISyncAPIClient',
  'MockOpenAIAsyncAPIClient',
]



from logging import Logger
from typing import Any, Optional
from munch import Munch
from faker import Faker
from llm_mocks.utils import serialise_to_json, write_to_json
import json
import os

LLM_ROLES = {
  "assistant": "assistant",
  "user": "user",
}


class DataFactory(object):
    def __init__(self, faker: Optional[Faker] = None, random=False):
        self.faker = faker
        self.random = random

    def get_object_data(self, data, override=None):
        if override:
            data.update(override)
        return Munch.fromDict(data)

    @staticmethod
    def load_default_data(file: str, name: str):
        path = os.path.join(os.path.dirname(file), 'data', f"{name}.json")
        with open(path) as f:
            data = json.load(f)
        return data

    def get_stream(self, override=None):
        pass

    def get(self, override=None):
        pass


class MockAPIClient():
    def __init__(
        self,
        *args,
        faker: Optional[Faker] = None,
        debug=False,
        **kwargs
    ):
        self.faker = faker
        self.debug = debug

    def log_function_call(self, logger: Logger, args, kwargs):
        if self.debug:
            logger.info(
                '#post &args: %s &kwargs: %s',
                ''.join('%s, ' % arg for arg in args),
                ''.join('%s=%s, ' % arg for arg in kwargs.items())
            )

    def get_json_path(self, path: Optional[str], module_name: str = '', is_stream: bool = False):
        if path is not None:
            if not os.path.exists(path):
                os.makedirs(path)
            file = os.path.join(path, f"{module_name}{'Stream' if is_stream else ''}.json")
            return file

    def record_stream_output(self, result, file):
        try:
            data = []
            for chunk in result:
                data.append(chunk)
                yield chunk
        finally:
            self.record_output(data, file)

    async def record_async_stream_output(self, result, file):
        try:
            data = []
            async for chunk in result:
                data.append(chunk)
                yield chunk
        finally:
            self.record_output(data, file)

    def record_output(self, result: Any, file: Optional[str]):
        if file is not None:
            with open(file, 'w') as f:
                write_to_json(result, f, ensure_ascii=False, indent=4)
        else:
            print(serialise_to_json(result))


__all__ = [
  'DataFactory',
  'MockAPIClient',
]



ANTHROPIC_MODELS = {
  "claude-3.5-Sonnet": "claude-3-5-sonnet-20240620",
  "claude-3-Sonnet": "claude-3-sonnet-20240229",
  "claude-3-Opus": "claude-3-opus-20240229",
  "claude-3-Haiku": "claude-3-haiku-20240307",
}



from llm_mocks.llm.base import DataFactory, LLM_ROLES
from .constants import OPEN_AI_CHAT_MODELS


MODULE_NAME = "ChatCompletion"


class OpenAIChatCompletionFactory(DataFactory):
    """
    Data Factory to return OpenAI ChatCompletion response in stream or normal format
    """
    name = MODULE_NAME
    stream_data = DataFactory.load_default_data(__file__, f"{MODULE_NAME}Stream")
    data = DataFactory.load_default_data(__file__, MODULE_NAME)

    def random_usage(self):
        if self.faker:
            promt_tokens = self.faker.random_int()
            completion_tokens = self.faker.random_int()
            return {
                "prompt_tokens": promt_tokens,
                "completion_tokens": completion_tokens,
                "total_tokens": promt_tokens + completion_tokens,
            }
        return {}

    def get_stream(self, override=None):
        data = OpenAIChatCompletionFactory.stream_data
        if self.faker:
            model = self.faker.random_choices(OPEN_AI_CHAT_MODELS)
            obj = "chat.completion.chunk"
            data = [
                {
                    "id": self.faker.uuid4(),
                    "model": model,
                    "object": obj,
                    "choices": [
                        {
                            "delta": {
                                "role": LLM_ROLES.get('assistant'),
                                "content": self.faker.text(),
                            },
                            "index": 0,
                        }
                    ]
                },
                {
                    "id": self.faker.uuid4(),
                    "model": model,
                    "object": obj,
                    "choices": [
                        {
                            "delta": {
                                "content": self.faker.text(),
                            },
                            "index": 0,
                        }
                    ]
                },
                {
                    "id": self.faker.uuid4(),
                    "model": model,
                    "object": obj,
                    "choices": [
                        {
                            "delta": {},
                            "finish_reason": "stop",
                            "index": 0,
                        }
                    ]
                },
            ]
        return self.get_object_data(data, override=override)

    def get(self, override=None):
        data = OpenAIChatCompletionFactory.data
        if self.faker:
            data = {
                "id": self.faker.uuid4(),
                "model": self.faker.random_choices(OPEN_AI_CHAT_MODELS),
                "object": "chat.completion",
                "usage": {
                    **self.random_usage(),
                },
                "choices": [
                    {
                        "finish_reason": "stop",
                        "index": 0,
                        "message": {
                            "id": self.faker.uuid4(),
                            "role": self.faker.random_choices(LLM_ROLES),
                            "content": self.faker.text()
                        }
                    }
                ]
            }

        return self.get_object_data(data, override=override)



from llm_mocks.llm.base import DataFactory


MODULE_NAME = "ImagesResponse"


class OpenAIImagesResponseFactory(DataFactory):
    """
    Data Factory to return OpenAI ImagesResponse response in stream or normal format
    """
    name = "ImagesResponse"
    data = DataFactory.load_default_data(__file__, MODULE_NAME)

    def get(self, override=None):
        data = OpenAIImagesResponseFactory.data
        if self.faker:
            data = {
                "data": [
                  {
                    "url": self.faker.url(),
                    "revised_prompt": self.faker.text()
                  }
                ]
            }

        return self.get_object_data(data, override=override)



from .llm import (
  MockAnthropicSyncAPIClient,
  AnthropicMessageFactory,
  OpenAIChatCompletionFactory,
  OpenAIImagesResponseFactory,
  MockOpenAISyncAPIClient,
  MockOpenAIAsyncAPIClient,
)


__all__ = [
  'MockAnthropicSyncAPIClient',
  'AnthropicMessageFactory',
  'OpenAIChatCompletionFactory',
  'OpenAIImagesResponseFactory',
  'MockOpenAISyncAPIClient',
  'MockOpenAIAsyncAPIClient',
]



OPEN_AI_CHAT_MODELS = {
  "gpt-4o": "gpt-4o",
  "gpt-4o-2024-05-13": "gpt-4o-2024-05-13",
  "gpt-4-turbo": "gpt-4-turbo",
  "gpt-4-turbo-2024-04-09": "gpt-4-turbo-2024-04-09",
  "gpt-4-turbo-preview": "gpt-4-turbo-preview",
  "gpt-4-0125-preview": "gpt-4-0125-preview",
  "gpt-4-1106-preview": "gpt-4-1106-preview",
  "gpt-4": "gpt-4",
  "gpt-4-0613": "gpt-4-0613",
  "gpt-4-0314": "gpt-4-0314",
  "gpt-3.5-turbo-0125": "gpt-3.5-turbo-0125",
  "gpt-3.5-turbo": "gpt-3.5-turbo",
  "gpt-3.5-turbo-1106": "gpt-3.5-turbo-1106",
  "gpt-3.5-turbo-instruct": "gpt-3.5-turbo-instruct"
}

OPEN_AI_IMAGE_MODELS = {
  "dall-e-3": "dall-e-3",
  "dall-e-2": "dall-e-2",
}

OPEN_AI_VOICE_MODELS = {
  "tts-1": "tts-1",
  "tts-1-hd": "tts-1-hd",
}

OPEN_AI_EMBEDDING_MODELS = {
  "text-embedding-3-large": "text-embedding-3-large",
  "text-embedding-3-small": "text-embedding-3-small",
  "text-embedding-ada-002": "text-embedding-ada-002",
}



from llm_mocks.llm.base import DataFactory, LLM_ROLES
from .constants import ANTHROPIC_MODELS


MODULE_NAME = "Message"


class AnthropicMessageFactory(DataFactory):
    """
    Data Factory to return Anthropic Message response in stream or normal format
    """
    name = MODULE_NAME
    stream_data = DataFactory.load_default_data(__file__, f"{MODULE_NAME}Stream")
    data = DataFactory.load_default_data(__file__, MODULE_NAME)

    def random_usage(self):
        if self.faker:
            return {
              "input_token": self.faker.random_int(),
              "output_token": self.faker.random_int(),
            }
        return {}

    def get_stream(self, override=None):
        data = AnthropicMessageFactory.stream_data
        if self.faker:
            model = self.faker.random_choices(ANTHROPIC_MODELS)
            role = self.faker.random_choices(LLM_ROLES)
            data = [
              {
                  "message": {
                      "id": self.faker.uuid4(),
                      "content": [],
                      "model": model,
                      "role": role,
                      "type": "message",
                      "usage": {
                          **self.random_usage()
                      }
                  },
                  "type": "message_start"
                },
              {
                  "delta": {
                      "text": self.faker.text(),
                      "type": "text_delta",
                  },
                  "index": 0,
                  "type": "content_block_delta"
              },
              {
                  "index": 0,
                  "type": "content_block_stop"
              },
              {
                  "delta": {
                      "stop_reason": "end_turn"
                  },
                  "type": "message_delta",
                  "usage": {
                      "output_tokens": self.faker.random_int(),
                  }
              },
              {
                  "type": "message_stop"
              }
            ]
        return self.get_object_data(data, override=override)

    def get(self, override=None):
        data = AnthropicMessageFactory.data
        if self.faker:
            data = {
                "id": self.faker.uuid4(),
                "model": self.faker.random_choices(ANTHROPIC_MODELS),
                "role": self.faker.random_choices(LLM_ROLES),
                "stop_reason": "end_turn",
                "content": [{
                    "text": self.faker.text(),
                    "type": "",
                }],
                "usage": {
                    **self.random_usage(),
                }
            }

        return self.get_object_data(data, override=override)



from typing import Callable
import logging
from llm_mocks.llm.base import MockAPIClient
from wrapt import wrap_function_wrapper
from .chat_completion import OpenAIChatCompletionFactory
from .images_response import OpenAIImagesResponseFactory

logger = logging.getLogger(__name__)


class MockOpenAIClient(MockAPIClient):
    """
    Base APIClient mock class. Required concret implementation
    """
    _base_client = ''
    _request_method = ''
    _request_path = f'{_base_client}.{_request_method}'

    def __init__(
        self,
        *args,
        ChatCompletionFactory=OpenAIChatCompletionFactory,
        ImagesResponseFactory=OpenAIImagesResponseFactory,
        **kwargs
    ):
        super().__init__(self, *args, **kwargs)
        self.chat_completion_factory = ChatCompletionFactory(faker=self.faker)
        self.images_response_factory = ImagesResponseFactory(faker=self.faker)

    def is_stream(self, kwargs):
        return kwargs.get('stream') is True

    def _request(self, *args, **kwargs):
        """
        Override this method to return correct mock response
        """
        pass

    def traced_method(self, *args, **kwargs):
        """
        Override this method to monkey patch method
        """
        pass

    def patch(self, patch: Callable):
        """
        Substitue original method with new mock implementation
        """
        patch(self._request_path, self._request)

    def record(self, path):
        """
        Wrap original method and print/write response output to json
        """
        wrap_function_wrapper(
          self._base_client,
          self._request_method,
          self.traced_method(path),
        )


class MockOpenAISyncAPIClient(MockOpenAIClient):
    """
    Monkey patch SyncAPIClient post to return or record mock response
    """
    _base_client = 'openai._base_client'
    _request_method = 'SyncAPIClient._request'
    _request_path = f'{_base_client}.{_request_method}'

    def _request(self, *args, **kwargs):
        self.log_function_call(logger, args, kwargs)

        request_name = kwargs.get('cast_to').__name__
        if self.is_stream(kwargs):
            if request_name == self.chat_completion_factory.name:
                def iter():
                    for item in self.chat_completion_factory.get_stream():
                        yield item
                return iter()

            return []
        else:
            if request_name == self.images_response_factory.name:
                return self.images_response_factory.get()

            if request_name == self.chat_completion_factory.name:
                return self.chat_completion_factory.get()

            return {}

    def traced_method(self, path=None):
        def wrapped(wrapped, instance, args, kwargs):
            is_stream = self.is_stream(kwargs)
            request_name = kwargs.get('cast_to').__name__
            file = self.get_json_path(path, module_name=request_name, is_stream=is_stream)
            result = wrapped(*args, **kwargs)

            if is_stream:
                return self.record_stream_output(result, file)
            else:
                self.record_output(result, file)
                return result

        return wrapped


class MockOpenAIAsyncAPIClient(MockOpenAIClient):
    """
    Monkey patch AsyncAPIClient post to return or record mock response
    """
    _base_client = 'openai._base_client'
    _request_method = 'AsyncAPIClient._request'
    _request_path = f'{_base_client}.{_request_method}'

    async def _request(self, *args, **kwargs):
        self.log_function_call(logger, args, kwargs)

        request_name = kwargs.get('cast_to').__name__
        if self.is_stream(kwargs):
            if request_name == self.chat_completion_factory.name:
                async def aiter():
                    for item in self.chat_completion_factory.get_stream():
                        yield item
                return aiter()

            return []
        else:
            if request_name == self.images_response_factory.name:
                return self.images_response_factory.get()

            if request_name == self.chat_completion_factory.name:
                return self.chat_completion_factory.get()

            return {}

    def traced_method(self, path=None):
        async def wrapped(wrapped, instance, args, kwargs):
            is_stream = self.is_stream(kwargs)
            request_name = kwargs.get('cast_to').__name__
            file = self.get_json_path(path, module_name=request_name, is_stream=is_stream)
            result = await wrapped(*args, **kwargs)

            if is_stream:
                return self.record_async_stream_output(result, file)
            else:
                self.record_output(result, file)
                return result

        return wrapped
